
from argparse import ArgumentParser, FileType
from time import time
import os

import numpy as np
from scipy.ndimage import laplace

from powerfit_em import Volume, Structure, quat_to_rotmat, proportional_orientations, determine_core_indices
from powerfit_em._powerfit import rotate_grid
from powerfit_em.volume import zeros_like, res_to_sigma, structure_to_shape


def parse_args():
    p = ArgumentParser(description='Perform a full rotational search of a structure in a density.')

    p.add_argument('target', type=FileType('rb'), help="Target density map.")
    p.add_argument('resolution', type=float, help="Resolution of target map")
    p.add_argument('template', type=FileType('r'), help="PDB file of template")

    p.add_argument('-a', '--angle', type=float, default=360, help="Rotational sampling interval in degree")
    p.add_argument('-l', '--laplace', dest='laplace', action='store_true', help="Use Laplacian pre-filter.")
    p.add_argument('-cw', '--core-weighted', dest='core_weighted', action='store_true', help='Use core-weighted correlation function')
    p.add_argument('-n', '--nsolutions', dest='nsolutions', type=int, default=2, help="Number of solutions written to file.")
    p.add_argument('-d', '--directory', dest='directory', type=str, default='.', help="Directory where results are stored.")
    p.add_argument('-o', '--output', dest='outfile', type=str, default='solutions.out', help="Filename holding the results.")

    args = p.parse_args()
    return args


def calc_lcc(target, template, mask, N):
    ind = mask != 0
    tmp = target * mask
    gcc = (tmp[ind] * template[ind]).sum()
    std = tmp[ind].std()
    lcc = gcc / std / N
    return lcc


def main():

    args = parse_args()
    try:
        os.makedirs(args.directory)
    except OSError:
        pass

    target = Volume.fromfile(args.target)
    structure = Structure.fromfile(args.template)
    center = structure.coor.mean(axis=1)
    radius = np.linalg.norm((structure.coor - center.reshape(-1, 1)), axis=0).max() + 0.5 * args.resolution

    template = zeros_like(target)
    rottemplate = zeros_like(target)
    mask = zeros_like(target)
    rotmask = zeros_like(target)
    structure_to_shape(structure.coor, args.resolution, out=template, shape='vol', weights=structure.atomnumber)
    structure_to_shape(structure.coor, args.resolution, out=mask, shape='mask')

    if args.laplace:
        target.array = laplace(target.array, mode='constant')
        template.array = laplace(template.array, mode='constant')
    if args.core_weighted:
        mask.array = determine_core_indices(mask.array)

    # Normalize the template density
    ind = mask.array != 0
    N = ind.sum()
    template.array *= mask.array
    template.array[ind] -= template.array[ind].mean()
    template.array[ind] /= template.array[ind].std()

    rotmat = quat_to_rotmat(proportional_orientations(args.angle)[0])

    lcc_list = []
    center -= target.origin
    center /= template.voxelspacing
    radius /= template.voxelspacing
    time0 = time()
    for n, rot in enumerate(rotmat):
        rotate_grid(template.array, rot, center, radius, rottemplate.array)
        rotate_grid(mask.array, rot, center, radius, rotmask.array, nearest=True)
        lcc = calc_lcc(target.array, rottemplate.array, rotmask.array, N)
        lcc_list.append(lcc)
        print('{:d}              \r'.format(n), end=' ')

    print('Searching took: {:.0f}m {:.0f}s'.format(*divmod(time() - time0, 60)))
    ind = np.argsort(lcc_list)[::-1]
    with open(os.path.join(args.directory, args.outfile), 'w') as f:
        line = ' '.join(['{:.4f}'] + ['{:7.4f}'] * 9) + '\n'
        for n in range(min(args.nsolutions, len(lcc_list))):
            f.write(line.format(lcc_list[ind[n]], *rotmat[ind[n]].ravel()))



if __name__ == "__main__":
    main()



import unittest

import numpy as np
from scipy.ndimage import laplace

try:
    import pyfftw as _
    PYFFTW = True
except ImportError:
    PYFFTW = False
try:
    import pyopencl as cl
    import pyopencl.array as cl_array

    OPENCL = True
except ImportError:
    OPENCL = False

from powerfit_em.powerfitter import CPUCorrelator
from powerfit_em.rotations import euler

if OPENCL:
    from powerfit_em.correlators.clkernels import CLKernels


@unittest.skipIf(not OPENCL, "GPU resources are not available.")
class TestCLKernels(unittest.TestCase):
    """Tests for the OpenCL kernels"""

    # @classmethod
    # def setUpClass(cls):
    #    p = cl.get_platforms()[0]
    #    devs = p.get_devices()
    #    cls.ctx = cl.Context(devices=devs)
    #    cls.queue = cl.CommandQueue(cls.ctx, device=devs[0])
    #    cls.k = CLKernels(cls.ctx)
    #    cls.s_linear = cl.Sampler(cls.ctx, False, cl.addressing_mode.CLAMP,
    #            cl.filter_mode.LINEAR)
    #    cls.s_nearest = cl.Sampler(cls.ctx, False, cl.addressing_mode.CLAMP,
    #            cl.filter_mode.NEAREST)

    def setUp(self):
        p = cl.get_platforms()[0]
        devs = p.get_devices()
        self.ctx = cl.Context(devices=devs)
        self.queue = cl.CommandQueue(self.ctx, device=devs[0])
        values = {
            "shape_x": 10,
            "shape_y": 0,
            "shape_z": 0,
            "llength": 5,
        }
        self.k = CLKernels(self.ctx, values=values)
        self.s_linear = cl.Sampler(
            self.ctx, False, cl.addressing_mode.CLAMP, cl.filter_mode.LINEAR
        )
        self.s_nearest = cl.Sampler(
            self.ctx, False, cl.addressing_mode.CLAMP, cl.filter_mode.NEAREST
        )

    def test_multiply(self):
        np_in1 = np.arange(10, dtype=np.float32)
        np_in2 = np.arange(10, dtype=np.float32)
        np_out = np_in1 * np_in2

        cl_in1 = cl_array.to_device(self.queue, np_in1)
        cl_out = cl_array.to_device(self.queue, np.zeros(10, dtype=np.float32))
        cl_in2 = cl_array.to_device(self.queue, np_in2)

        self.k.multiply(cl_in1, cl_in2, cl_out)
        self.assertTrue(np.allclose(np_out, cl_out.get()))

    def test_conj_multiply(self):
        np_in1 = np.zeros(10, dtype=np.complex64)
        np_in2 = np.zeros(10, dtype=np.complex64)
        np_in1.real = np.random.rand(10)
        np_in1.imag = np.random.rand(10)
        np_in2.real = np.random.rand(10)
        np_in2.imag = np.random.rand(10)
        np_out = np_in1.conj() * np_in2

        cl_in1 = cl_array.to_device(self.queue, np_in1)
        cl_in2 = cl_array.to_device(self.queue, np_in2)
        cl_out = cl_array.to_device(self.queue, np.zeros(10, dtype=np.complex64))
        self.k.conj_multiply(cl_in1, cl_in2, cl_out)
        self.assertTrue(np.allclose(np_out, cl_out.get()))

    # TODO
    # def test_calc_lcc(self):
    #    pass
    # def test_take_best(self):
    #    pass

    @unittest.skip("kernel does not have rotate_template")
    def test_rotate_template_mask(self):
        shape = (5, 5, 5)
        values = {
            "shape_x": 5,
            "shape_y": 5,
            "shape_z": 5,
            "llength": 2,
        }
        self.k = CLKernels(self.ctx, values=values)
        template = np.zeros(shape, dtype=np.float32)
        template[2, 2, 1:4] = 1
        template[2, 1:4, 2] = 1
        rotmat = np.asarray([1, 0, 0, 0, 1, 0, 0, 0, 1] + [0] * 7, dtype=np.float32)

        self.queue.finish()
        cl_template = cl.image_from_array(self.queue.context, template)
        cl_out = cl_array.to_device(self.queue, np.zeros(shape, dtype=np.float32))
        center = np.asarray([2, 2, 2, 0], dtype=np.float32)
        shape = np.asarray([5, 5, 5, 125], dtype=np.int32)

        self.k.rotate_template(
            self.queue,
            (125,),
            None,
            self.s_linear,
            cl_template,
            rotmat,
            cl_out.data,
            center,
            shape,
        )

        answer = np.zeros((5, 5, 5), dtype=np.float32)
        answer[0, 0, :2] = 1
        answer[0, 0, -1] = 1
        answer[0, :2, 0] = 1
        answer[0, -1, 0] = 1

        self.assertTrue(np.allclose(cl_out.get(), answer))

    @unittest.skip("kernel does not have rotate_grids_and_multiply")
    def test_rotate_grids_and_multiply(self):
        shape = (5, 5, 5)
        values = {
            "shape_x": 5,
            "shape_y": 5,
            "shape_z": 5,
            "llength": 2,
        }
        self.k = CLKernels(self.ctx, values=values)

        template = np.zeros(shape, dtype=np.float32)
        template[2, 2, 1:4] = 1
        template[2, 1:4, 2] = 1
        mask = template * 2
        np_out_template = np.zeros(shape, dtype=np.float32)
        np_out_template[0, 0, :2] = 1
        np_out_template[0, 0, -1] = 1
        np_out_template[0, :2, 0] = 1
        np_out_template[0, -1, 0] = 1
        np_out_mask = np_out_template * 2
        np_out_mask2 = np_out_mask**2

        cl_template = cl.image_from_array(self.ctx, template)
        cl_mask = cl.image_from_array(self.ctx, mask)
        cl_rotmat = np.asarray([1, 0, 0, 0, 1, 0, 0, 0, 1] + [0] * 7, dtype=np.float32)
        cl_center = np.asarray([2, 2, 2, 0], dtype=np.float32)
        cl_shape = np.asarray([5, 5, 5, 125], dtype=np.int32)
        cl_radius = np.int32(2)

        cl_out_template = cl_array.to_device(
            self.queue, np.zeros(shape, dtype=np.float32)
        )
        cl_out_mask = cl_array.to_device(self.queue, np.zeros(shape, dtype=np.float32))
        cl_out_mask2 = cl_array.to_device(self.queue, np.zeros(shape, dtype=np.float32))

        gws = tuple([int(2 * cl_radius + 1)] * 3)
        args = (
            cl_template,
            cl_mask,
            cl_rotmat,
            self.s_linear,
            self.s_nearest,
            cl_center,
            cl_shape,
            cl_radius,
            cl_out_template.data,
            cl_out_mask.data,
            cl_out_mask2.data,
        )
        self.k.rotate_grids_and_multiply(self.queue, gws, None, *args)
        self.queue.finish()

        self.assertTrue(np.allclose(np_out_template, cl_out_template.get()))
        self.assertTrue(np.allclose(np_out_mask, cl_out_mask.get()))
        self.assertTrue(np.allclose(np_out_mask2, cl_out_mask2.get()))


if __name__ == "__main__":
    unittest.main()



import numpy as np
import pyopencl as cl
from pyopencl.array import Array as ClArray
from pyopencl.elementwise import ElementwiseKernel


from string import Template
import importlib.resources

import powerfit_em.correlators


class CLKernels(object):
    def __init__(self, ctx, values):
        self.sampler_nearest = cl.Sampler(ctx, True,
                cl.addressing_mode.REPEAT, cl.filter_mode.NEAREST)
        self.sampler_linear = cl.Sampler(ctx, True,
                cl.addressing_mode.REPEAT, cl.filter_mode.LINEAR)
        self.multiply = ElementwiseKernel(ctx,
                "float *x, float *y, float *z",
                "z[i] = x[i] * y[i];"
                )
        self.conj_multiply = ElementwiseKernel(ctx,
                "cfloat_t *x, cfloat_t *y, cfloat_t *z",
                "z[i] = cfloat_mul(cfloat_conj(x[i]), y[i]);"
                )
        self.calc_lcc_and_take_best = ElementwiseKernel(ctx,
                """float *gcc, float *ave, float *ave2, int *mask,
                    float norm_factor, int nrot, float *lcc, int *grot""",
                """float _lcc;
                    if (mask[i] > 0) {
                        _lcc = gcc[i] / sqrt(ave2[i] * norm_factor - ave[i] * ave[i]);
                        if (_lcc > lcc[i]) {
                            lcc[i] = _lcc;
                            grot[i] = nrot;
                        };
                    };
                """
                )

        t = importlib.resources.read_text(powerfit_em.correlators, "kernels.cl")
        t = Template(t).substitute(**values)

        self._program = cl.Program(ctx, t).build()
        self._rotate_image3d = self._program.rotate_image3d
        self._gws_rotate_grid3d = (96, 64, 1)

    def rotate_grid3d(
        self,
        queue: cl.CommandQueue,
        grid:ClArray,
        rotmat: np.ndarray,
        out: ClArray,
        nearest: bool = False
    ):
        args = (grid.data, rotmat, out.data, np.int32(nearest))
        self._program.rotate_grid3d(queue, self._gws_rotate_grid3d, None, *args)

    def rotate_image3d(self, queue: cl.CommandQueue, image, rotmat, out, nearest=False):
        if nearest:
            args = (image, self.sampler_nearest, rotmat, out.data)
        else:
            args = (image, self.sampler_linear, rotmat, out.data)
        self._rotate_image3d(queue, self._gws_rotate_grid3d, None, *args)




from functools import partial
from os import remove
from os.path import join, abspath, isdir
from multiprocessing import RawValue, Lock, Process

import numpy as np
from tqdm.auto import tqdm

from powerfit_em.correlators.cpu import CPUCorrelator
from powerfit_em.volume import Volume
try:
    import pyfftw as _
    PYFFTW = True
except ImportError:
    PYFFTW = False
try:
    import pyopencl as _
    OPENCL = True
except:
    OPENCL = False


class _Counter(object):
    """Thread-safe counter object to follow PowerFit progress"""

    def __init__(self):
        self.val = RawValue('i', 0)
        self.lock = Lock()

    def increment(self):
        with self.lock:
            self.val.value += 1

    def value(self):
        with self.lock:
            return self.val.value


def run_correlator_instance(
    target: Volume,
    template: Volume,
    mask: Volume,
    rotations: np.ndarray,
    laplace: bool,
    jobid: int,
    directory: str,
    counter: _Counter | None
):
    correlator = CPUCorrelator(
        target.array,
        template.array,
        rotations,
        mask.array,
        laplace,
    )
    for n in range(len(rotations)):
        correlator.compute_rotation(n, rotmat=correlator.rotations[n])
        if counter is not None:
            counter.increment()

    np.save(join(directory, '_lcc_part_{:d}.npy').format(jobid), correlator.lcc)
    np.save(join(directory, '_rot_part_{:d}.npy').format(jobid), correlator.rot)


class PowerFitter(object):
    """Wrapper around the Correlator classes for multiprocessing and GPU
    accelerated searches providing an easy interface.
    """

    def __init__(
        self, target: Volume, rotations: np.ndarray, template: Volume, mask: Volume, queues, laplace: bool = False
    ):
        self._target = target
        self._rotations = rotations
        self._template = template
        self._mask = mask
        self._queues = queues
        self._nproc = 1
        self._directory = abspath('./')
        self._laplace = laplace

    @property
    def directory(self):
        return self._directory

    @directory.setter
    def directory(self, directory):
        if isdir(directory):
            self._directory = abspath(directory)
        else:
            raise ValueError("Directory does not exist.")

    def scan(self, progress: partial[tqdm] | None):
        if self._queues is None:
            if self._nproc == 1:
                self._single_cpu_scan(progress)
            else:
                self._multi_cpu_scan(progress)
        else:
            self._gpu_scan(progress)

    def _gpu_scan(self, progress: partial[tqdm] | None):
        if OPENCL:
            from powerfit_em.correlators.gpu import GPUCorrelator
        self._corr = GPUCorrelator(
            self._target.array,
            self._template.array,
            self._rotations,
            self._mask.array,
            self._queues[0],
            self._laplace,
        )
        self._corr.scan(progress)
        self._lcc = self._corr.lcc
        self._rot = self._corr.rot

    def _multi_cpu_scan(self, progress: partial[tqdm] | None):
        nrot = self._rotations.shape[0]
        self._nrot_per_job = nrot // self._nproc
        processes: list[Process] = []
        self._counter = None if processes is None else _Counter()
        self._njobs = self._nproc
        if self._queues is not None:
            self._njobs = len(self._queues)

        for id in range(self._njobs):
            start = id * self._nrot_per_job
            stop = start + self._nrot_per_job
            if id == self._njobs - 1:
                stop = len(self._rotations)
            partial_rotations = self._rotations[start:stop]
            processes.append(
                Process(
                  target=run_correlator_instance,
                  args=(self._target, self._template, self._mask,
                        partial_rotations, self._laplace, id,
                        self._directory, self._counter)
                )
            )

        for id in range(self._njobs):
            processes[id].start()

        if progress is not None:
            with progress(total=nrot) as pbar:
                while self._counter.value() < nrot:
                    current_count = self._counter.value()
                    pbar.update(current_count - pbar.n)
        
        for id in range(self._njobs):
            processes[id].join()
        self._combine()

    def _single_cpu_scan(self, progress: partial[tqdm] | None):
        correlator = CPUCorrelator(
            self._target.array,
            self._template.array,
            self._rotations,
            self._mask.array,
            self._laplace,
        )
        correlator.scan(progress)
        self._lcc = correlator.lcc
        self._rot = correlator.rot

    def _combine(self):
        # Combine all the intermediate results
        lcc = np.zeros(self._target.shape)
        rot = np.zeros(self._target.shape)
        ind = np.zeros(lcc.shape, dtype=np.bool)
        for n in range(self._njobs):
            lcc_file = join(self._directory, '_lcc_part_{:d}.npy').format(n)
            rot_file = join(self._directory, '_rot_part_{:d}.npy').format(n)
            part_lcc = np.load(lcc_file)
            part_rot = np.load(rot_file)
            np.greater(part_lcc, lcc, ind)
            lcc[ind] = part_lcc[ind]
            # take care of the rotation index offset for each independent job
            rot[ind] = part_rot[ind] + self._nrot_per_job * n
            remove(lcc_file)
            remove(rot_file)
        self._lcc = lcc
        self._rot = rot



from numpy import zeros, bool, greater_equal, log
from scipy.ndimage import label, maximum_position

class Analyzer(object):


    def __init__(self, corr, rotmat, rotmat_ind, steps=5, voxelspacing=1,
                 origin=(0, 0, 0), z_sigma=1):
        self._corr = corr
        self._rotmat = rotmat
        self._rotmat_ind = rotmat_ind
        self._voxelspacing = voxelspacing
        self._origin = origin
        self._z_sigma = z_sigma
        self.steps = steps
        self._solutions = None

    @property
    def corr(self):
        return self._corr

    @property
    def rot(self):
        return self._rot

    @property
    def steps(self):
        return self._steps

    @steps.setter
    def steps(self, steps):
        self._steps = steps
        self._watershed()
        self._solutions = None

    @property
    def voxelspacing(self):
        return self._voxelspacing

    @voxelspacing.setter
    def voxelspacing(self, voxelspacing):
        self._solutions = None
        self._voxelspacing = voxelspacing

    @property
    def origin(self):
        return self._origin

    @origin.setter
    def origin(self, origin):
        self._solutions = None
        self._origin = origin

    @property
    def solutions(self):
        if self._solutions is None:
            self._generate_solutions()
        return self._solutions

    def _generate_solutions(self):
        solutions = []
        for pos in self._positions:
            solution = []
            # To support scipy 0.14.0 cast pos into tuple of ints since
            # maximum_position returns a list of tuples of floats
            pos = tuple(int(p) for p in pos)
            lcc = self._corr[pos]
            solution.append(lcc)
            fishers_z = 0.5 * (log(1 + lcc) - log(1 - lcc))
            solution.append(fishers_z)
            rel_z = fishers_z / self._z_sigma
            solution.append(rel_z)
            z, y, x = [coor * self._voxelspacing  + shift for coor, shift in
                    zip(pos, self._origin[::-1])]
            rotmat = self._rotmat[int(self._rotmat_ind[pos])]
            solution += [x, y, z] + list(rotmat.ravel())
            solutions.append(solution)
        self._solutions = sorted(solutions, key=lambda cc: cc[0], reverse=True)

    def _watershed(self):
        """Get positions of high correlation values using watershed
        algorithm
        """
        max_cc = self._corr.max()
        min_cc = 0.5 * max_cc
        stepsize = (max_cc - min_cc) / self._steps
        cutoff = max_cc
        positions = []
        mask = zeros(self._corr.shape, dtype=bool)
        for n in range(self._steps):
            cutoff -= stepsize
            greater_equal(self._corr, cutoff, mask)
            labels, nfeatures = label(mask)
            positions += maximum_position(self._corr, labels, list(range(1, nfeatures + 1)))
        self._positions = set(positions)

    def tofile(self, out='solutions.out', delimiter=None):
        """Write solutions to file.

        Arguments:
            out: str, output file name (default: 'solutions.out')
            delimiter: str, delimiter for the output file. Defaults to fixed width output.
                Can be ',' or '\t'. With delimiter set the header will not have leading '#'.
        """

        if self._solutions is None:
            self._generate_solutions()

        headers = '#rank cc Fish-z rel-z x y z a11 a12 a13 a21 a22 a23 a31 a32 a33'.split()
        with open(out, 'w') as f:
            if delimiter is None:
                line = ' '.join(['{:<6s}'] + ['{:>6s}'] * 3 + ['{:>8s}'] * 3 + ['{:>6s}'] * 9) + '\n'
                f.write(line.format(*headers))
                line = ' '.join(['{:<6d}'] + ['{:6.3f}'] * 3 + ['{:8.3f}'] * 3 + ['{:6.3f}'] * 9) + '\n'
                for n, sol in enumerate(self._solutions):
                    f.write(line.format(n + 1, *sol))
            else:
                # Write header
                headers[0] = headers[0].lstrip('#')
                f.write(delimiter.join(headers) + '\n')
                for n, sol in enumerate(self._solutions):
                    row = [n + 1] + sol
                    # Format all values as floats with 3 decimals except rank (int)
                    row_str = [str(row[0])] + [f"{v:.3f}" for v in row[1:4]] + [f"{v:.3f}" for v in row[4:7]] + [f"{v:.3f}" for v in row[7:]]
                    f.write(delimiter.join(row_str) + '\n')






import os
import errno
from math import sqrt, pi

import numpy as np
from scipy.ndimage import binary_erosion
try:
    import pyopencl as cl
except ImportError:
    pass

from . import volume


def mkdir_p(path):
    try:
        os.makedirs(path)
    except OSError as exc:
        if exc.errno == errno.EEXIST and os.path.isdir(path):
            pass
        else:
            raise


def determine_core_indices(mask):
    """Calculate the core indices of a shape"""

    core_indices = np.zeros(mask.shape)
    eroded_mask = mask > 0
    while eroded_mask.sum() > 0:
        core_indices += eroded_mask
        eroded_mask = binary_erosion(eroded_mask)
    return core_indices


def get_queue(platformid=0, deviceid=0):
    try:
        platform = cl.get_platforms()[platformid]
        devices = platform.get_devices()
        context = cl.Context(devices=devices)
        queue = cl.CommandQueue(context, device=devices[deviceid])
    except Exception as e:
        raise e
        queue = None

    return queue


def fisher_sigma(mv, fsc):
    return 1 / sqrt(mv / fsc - 3)


def write_fits_to_pdb(structure, solutions, basename='fit'):
    translated_structure = structure.duplicate()
    center = translated_structure.coor.mean(axis=1)
    translated_structure.translate(-center)
    for n, sol in enumerate(solutions, start=1):
        out = translated_structure.duplicate()
        rot = np.asarray([float(x) for x in sol[6:]]).reshape(3, 3)
        trans = sol[3:6]
        out.rotate(rot)
        out.translate(trans)
        out.tofile(basename + '_{:d}.pdb'.format(n))




from argparse import ArgumentParser, FileType
from os.path import splitext

from powerfit_em.volume import Volume, trim, resample, lower_resolution

EM2EM_DESCRIPTION = "Convert, trim, and resample cryo-EM data."


def parse_em2em():
    p = ArgumentParser(description=EM2EM_DESCRIPTION)
    p.add_argument('infile', type=str,
            help="Input cryo-EM density file.")
    p.add_argument("outfile", type=str,
            help="Output cryo-EM density file.")

    p.add_argument('-f1', '--format-input', default=None,
            dest='input_format', metavar="<str>",
            help="Format of input file.")
    p.add_argument('-f2', '--format-output', default=None,
            dest='output_format', metavar="<str>",
            help="Format of output file.")
    p.add_argument("-t", "--trim", action='store_true', dest="trim",
            help="Trim the density.")
    p.add_argument('-tc', "--trim-cutoff", type=float, default=None,
            dest="trim_cutoff",
            help="Cutoff value for trimming.")
    p.add_argument("-r", "--resample", default=None, type=float,
            dest="resample", metavar='<float>',
            help="Resample the density to a specified voxel spacing in angstrom.")

    args = p.parse_args()
    return args


def em2em():
    args = parse_em2em()

    print('Reading input file ...')
    v = Volume.fromfile(args.infile, args.input_format)

    if args.resample is not None:
        print('Resampling ...')
        v = resample(v, v.voxelspacing / args.resample)

    if args.trim:
        print('Trimming ...')
        if args.trim_cutoff is None:
            args.trim_cutoff = 0.1 * v.array.max()
        v = trim(v, args.trim_cutoff)

    print('Writing to file ...')
    v.tofile(args.outfile, args.output_format)


def parse_image_pyramid():
    p = ArgumentParser()

    p.add_argument('map', type=FileType('rb'), 
            help='Initial density data.')
    p.add_argument('resolution', type=float, 
            help='Resolution of initial data.')
    p.add_argument('target_resolutions', nargs='+', type=float,
            help='The target resolutions of the resulting image-pyramid.')

    p.add_argument('-rr', '--resampling-rate', dest='resampling_rate',
            type=float, default=2, metavar='<float>',
            help=('Nyquist resampling rate. Default is 2 x '
            'Nyquist, i.e. resulting voxelspacing is '
            '1/4th of the resolution.'),
            )
    p.add_argument('-b', '--base-name', dest='base_name', type=str, default=None,
            metavar='<string>',
            help='Base name of the resulting maps. Default is original mapfile name.')
    args = p.parse_args()

    # some error checking
    if args.resolution <= 0:
        raise ValueError('Resolution should be bigger than 0.')
    if args.resampling_rate < 1:
        raise ValueError('Resampling rate should be bigger than 1 times Nyquist.')
    for resolution in args.target_resolutions:
        if resolution < args.resolution:
            raise ValueError('Target resolution of image-pyramid should be '
                'lower than original data.')
    if args.base_name is None:
        args.base_name = splitext(args.map.name)[0] 

    return args


def image_pyramid():
    args = parse_image_pyramid()

    # create image-pyramid
    vol = Volume.fromfile(args.map)
    fname = args.base_name + '_{:.0f}.mrc'
    for resolution in args.target_resolutions:
        vol2 = lower_resolution(vol, args.resolution, resolution)
        new_voxelspacing = resolution / (2 * args.resampling_rate)
        factor = vol.voxelspacing / new_voxelspacing 
        vol2 = resample(vol2, factor, order=1)
        vol2.tofile(fname.format(resolution))




__version__ = "3.0.6"

from .volume import Volume, structure_to_shape_like
from .structure import Structure
from .rotations import proportional_orientations, quat_to_rotmat
from .helpers import determine_core_indices



import sys

from powerfit_em.structure import Structure
from powerfit_em.volume import structure_to_shape


def main():

    structure = Structure.fromfile(sys.argv[1])
    template = structure_to_shape(
        structure.coor,
        resolution=float(sys.argv[2]),
        weights=structure.atomnumber,
        shape="vol",
    )
    template.tofile(sys.argv[3])


if __name__ == "__main__":
    main()



from unittest import TestCase, main, skipIf

import numpy as np
import numpy.testing as npt

from powerfit_em.helpers import get_queue
from powerfit_em._extensions import rotate_grid3d

try:
    from powerfit_em.powerfitter import CLKernels
    import pyopencl as cl
    import pyopencl.array as cl_array

    HAS_CL = True
except ImportError:
    HAS_CL = False


@skipIf(not HAS_CL, "CL not available")
class TestCLKernels(TestCase):

    @classmethod
    def setUpClass(self):

        self.queue = get_queue()
        self.shape = (4, 5, 6)
        self.size = 4 * 5 * 6
        self.values = {
            "shape_x": self.shape[2],
            "shape_y": self.shape[1],
            "shape_z": self.shape[0],
            "llength": 2,
        }
        self.k = CLKernels(self.queue.context, self.values)
        self.grid = np.zeros(self.shape, dtype=np.float64)
        self.grid[0, 0, 0] = 1
        self.grid[0, 0, 1] = 1
        self.grid[0, 1, 1] = 1
        self.grid[0, 0, 2] = 1
        self.grid[0, 0, -1] = 1
        self.grid[-1, 0, 0] = 1
        self.cl_image = cl.image_from_array(
            self.queue.context, self.grid.astype(np.float32)
        )
        self.sampler_linear = cl.Sampler(
            self.queue.context, True, cl.addressing_mode.REPEAT, cl.filter_mode.LINEAR
        )
        self.sampler_nearest = cl.Sampler(
            self.queue.context, False, cl.addressing_mode.CLAMP, cl.filter_mode.NEAREST
        )
        self.out = np.zeros(self.shape, dtype=np.float64)
        self.cl_out = cl_array.zeros(self.queue, self.shape, dtype=np.float32)

    def test_rotate_image3d_linear(self):
        k = self.k._program.rotate_image3d

        rotmat = np.asarray([[1, 0, 0], [0, 1, 0], [0, 0, 1]], dtype=np.float64)
        cl_rotmat = np.asarray([1, 0, 0, 0, 1, 0, 0, 0, 1] + [0] * 7, dtype=np.float32)
        args = (self.cl_image, self.sampler_linear, cl_rotmat, self.cl_out.data)
        k(self.queue, (5, 5, 5), None, *args)
        rotate_grid3d(self.grid, rotmat, 2, self.out, False)

        test = np.allclose(self.cl_out.get(), self.out)
        self.assertTrue(test)

        # 90 degree rotation around z-axis
        rotmat = np.asarray([[0, -1, 0], [1, 0, 0], [0, 0, 1]], dtype=np.float64)
        cl_rotmat = np.asarray([0, -1, 0, 1, 0, 0, 0, 0, 1] + [0] * 7, dtype=np.float32)
        args = (self.cl_image, self.sampler_linear, cl_rotmat, self.cl_out.data)
        k(self.queue, (5, 5, 5), None, *args)
        rotate_grid3d(self.grid, rotmat, 2, self.out, False)

        test = np.allclose(self.cl_out.get(), self.out)
        self.assertTrue(test)

        # Non-integer rotation
        rotmat = np.asarray(
            [
                [0.30901699, -0.5, 0.80901699],
                [-0.80901699, 0.30901699, 0.5],
                [-0.5, -0.80901699, -0.30901699],
            ],
            dtype=np.float64,
        )
        cl_rotmat = np.asarray(rotmat.ravel().tolist() + [0] * 7, dtype=np.float32)
        args = (self.cl_image, self.sampler_linear, cl_rotmat, self.cl_out.data)
        k(self.queue, (1, 1, 1), None, *args)
        rotate_grid3d(self.grid, rotmat, 2, self.out, False)
        test = np.allclose(self.cl_out.get(), self.out)
        # There will be difference between CPU and GPU, but the results should
        # be less than 5% different. This is an ad hoc number.
        diff = np.nan_to_num(np.abs((self.cl_out.get() - self.out) / self.out)).max()
        self.assertTrue(diff < 0.05)

    def test_rotate_grid3d_nearest(self):
        """Test rotate_grid3d kernel using nearest interpolation."""
        k = self.k._program.rotate_grid3d

        # Identity rotation
        rotmat = np.asarray([1, 0, 0, 0, 1, 0, 0, 0, 1] + [0] * 7, dtype=np.float32)
        grid = np.zeros((4, 5, 6), dtype=np.float32)
        grid[0, 0, 0] = 1
        grid[0, 0, 1] = 1
        grid[0, 1, 1] = 1
        grid[0, 0, 2] = 1
        grid[0, 0, -1] = 1
        grid[-1, 0, 0] = 1
        self.cl_grid = cl_array.to_device(self.queue, grid)
        self.cl_out = cl_array.zeros_like(self.cl_grid)

        args = (self.cl_grid.data, rotmat, self.cl_out.data, np.int32(True))
        gws = tuple([2 * self.values["llength"] + 1] * 3)
        k(self.queue, gws, None, *args)

        self.assertTrue(np.allclose(self.cl_grid.get(), self.cl_out.get()))

        # 90' rotation around z-axis
        self.cl_out.fill(0)
        rotmat = np.asarray([0, -1, 0, 1, 0, 0, 0, 0, 1] + [0] * 7, dtype=np.float32)
        args = (self.cl_grid.data, rotmat, self.cl_out.data, np.int32(True))
        gws = tuple([2 * self.values["llength"] + 1] * 3)
        k(self.queue, gws, None, *args)

        answer = np.zeros(self.shape, dtype=np.float32)
        answer[0, 0, 0] = 1
        answer[0, 1, 0] = 1
        answer[0, 1, -1] = 1
        answer[0, 2, 0] = 1
        answer[0, -1, 0] = 1
        answer[-1, 0, 0] = 1
        self.assertTrue(np.allclose(answer, self.cl_out.get()))

    @skipIf(
        True, "rotate_grid3d not executed by powerfit cli with --gpu, ignoring test"
    )
    def test_rotate_grid3d_linear(self):
        """Test rotate_grid3d kernel using nearest interpolation."""
        k = self.k._program.rotate_grid3d

        # Identity rotation
        rotmat = np.asarray([1, 0, 0, 0, 1, 0, 0, 0, 1] + [0] * 7, dtype=np.float32)
        grid = np.zeros((4, 5, 6), dtype=np.float32)
        grid[0, 0, 0] = 1
        grid[0, 0, 1] = 1
        grid[0, 1, 1] = 1
        grid[0, 0, 2] = 1
        grid[0, 0, -1] = 1
        grid[-1, 0, 0] = 1
        self.cl_grid = cl_array.to_device(self.queue, grid)
        self.cl_out = cl_array.zeros_like(self.cl_grid)

        args = (self.cl_grid.data, rotmat, self.cl_out.data, np.int32(False))
        gws = tuple([2 * self.values["llength"] + 1] * 3)
        k(self.queue, gws, None, *args)

        with open("cl_grid.txt", "w") as f:
            f.write(str(self.cl_grid.get()))
        with open("cl_out.txt", "w") as f:
            f.write(str(self.cl_out.get()))

        npt.assert_allclose(
            self.cl_grid.get(),
            self.cl_out.get(),
        )

        # 90' rotation around z-axis
        self.cl_out.fill(0)
        rotmat = np.asarray([0, -1, 0, 1, 0, 0, 0, 0, 1] + [0] * 7, dtype=np.float32)
        args = (self.cl_grid.data, rotmat, self.cl_out.data, np.int32(False))
        gws = tuple([2 * self.values["llength"] + 1] * 3)
        k(self.queue, gws, None, *args)

        answer = np.zeros(self.shape, dtype=np.float32)
        answer[0, 0, 0] = 1
        answer[0, 1, 0] = 1
        answer[0, 1, -1] = 1
        answer[0, 2, 0] = 1
        answer[0, -1, 0] = 1
        answer[-1, 0, 0] = 1
        npt.assert_allclose(answer, self.cl_out.get())

        # Non-integer rotation
        rotmat = np.asarray(
            [
                [0.30901699, -0.5, 0.80901699],
                [-0.80901699, 0.30901699, 0.5],
                [-0.5, -0.80901699, -0.30901699],
            ],
            dtype=np.float64,
        )
        cl_rotmat = np.asarray(rotmat.ravel().tolist() + [0] * 7, dtype=np.float32)
        args = (self.cl_grid.data, cl_rotmat, self.cl_out.data, np.int32(False))
        k(self.queue, gws, None, *args)
        rotate_grid3d(self.grid, rotmat, 2, self.out, False)
        npt.assert_allclose(self.cl_out.get(), self.out)

        # self.assertTrue(test)


if __name__ == "__main__":
    main()




from math import sqrt
from random import random
import importlib.resources

import numpy as np

import powerfit_em.data


def euler(angle, axis):
    cos_a = np.cos(angle)
    sin_a = np.sin(angle)
    if axis == 'x':
        matrix = [[1, 0, 0], [0, cos_a, -sin_a], [0, sin_a, cos_a]]
    elif axis == 'y':
        matrix = [[cos_a, 0, sin_a], [0, 1, 0], [-sin_a, 0, cos_a]]
    elif axis == 'z':
        matrix = [[cos_a, -sin_a, 0], [sin_a, cos_a, 0], [0, 0, 1]]
    else:
        raise ValueError('Axis is not recognized.')
    return np.asmatrix(matrix, dtype=np.float64)


def euler_to_rotmat(angles, order='zyz'):
    rotmat = np.asmatrix(np.identity(3, dtype=np.float64))
    for angle, axis in zip(angles, order):
        rotmat *= euler(angle, axis)
    return rotmat


def quat_to_rotmat(quaternions):

    w = quaternions[:, 0]
    x = quaternions[:, 1]
    y = quaternions[:, 2]
    z = quaternions[:, 3]

    Nq = w**2 + x**2 + y**2 + z**2
    s = np.zeros(Nq.shape, dtype=np.float64)
    s[Nq >  0.0] = 2.0/Nq[Nq > 0.0]
    s[Nq <= 0.0] = 0

    X = x*s
    Y = y*s
    Z = z*s

    rotmat = np.zeros((quaternions.shape[0],3,3), dtype=np.float64)
    rotmat[:,0,0] = 1.0 - (y*Y + z*Z)
    rotmat[:,0,1] = x*Y - w*Z
    rotmat[:,0,2] = x*Z + w*Y

    rotmat[:,1,0] = x*Y + w*Z
    rotmat[:,1,1] = 1.0 - (x*X + z*Z)
    rotmat[:,1,2] = y*Z - w*X

    rotmat[:,2,0] = x*Z - w*Y
    rotmat[:,2,1] = y*Z + w*X
    rotmat[:,2,2] = 1.0 - (x*X + y*Y)

    np.around(rotmat, decimals=8, out=rotmat)

    return rotmat


def random_rotmat():
    """Return a random rotation matrix"""

    s1 = 1
    while s1 >= 1.0:
        e1 = random() * 2 - 1
        e2 = random() * 2 - 1
        s1 = e1**2 + e2**2
            
    s2 = 1
    while s2 >= 1.0:
        e3 = random() * 2 - 1
        e4 = random() * 2 - 1
        s2 = e3**2 + e4**2
    
    q0 = e1
    q1 = e2
    q2 = e3 * sqrt((1 - s1)/s2 )
    q3 = e4 * sqrt((1 - s1)/s2 )

    quat = [[q0, q1, q2, q3]]
    return quat_to_rotmat(np.asarray(quat))[0]


def proportional_orientations(angle):
    # orientation sets available: name of file: (Norientations, degree)
    rot_sets = {'E.npy': (1, 360.0),
                'c48u1.npy': (24, 62.8),
                'c600v.npy': (60, 44.48),
                'c48n9.npy': (216, 36.47),
                'c600vc.npy': (360, 27.78),
                'c48u27.npy': (648, 20.83),
                'c48u83.npy': (1992, 16.29),
                'c48u181.npy': (4344, 12.29),
                'c48n309.npy': (7416, 9.72),
                'c48n527.npy': (12648, 8.17),
                'c48u815.npy': (19560, 7.4),
                'c48u1153.npy': (27672, 6.6),
                'c48u1201.npy': (28824, 6.48),
                'c48u1641.npy': (39384, 5.75),
                'c48u2219.npy': (53256, 5.27),
                'c48u2947.npy': (70728, 4.71),
                'c48u3733.npy': (89592, 4.37),
                'c48u4749.npy': (113976, 4.0),
                'c48u5879.npy': (141096, 3.74),
                'c48u7111.npy': (170664, 3.53),
                'c48u8649.npy': (207576, 3.26),
                }

    # determine the apropiate set to use
    smallestdiff = None
    for s, n in rot_sets.items():
        alpha = n[1]
        diff = abs(angle - alpha)
        if smallestdiff is None or diff < smallestdiff:
            smallestdiff = diff
            fname = s

    # read file
    infile = importlib.resources.open_binary(powerfit_em.data, fname)
    quat_weights = np.load(infile)

    quat = quat_weights[:, :4]
    weights = quat_weights[:, -1]
    alpha = rot_sets[fname][1]

    return quat, weights, alpha



import unittest
from unittest import skipIf

import numpy as np

from powerfit_em import _powerfit
from powerfit_em.rotations import euler
from powerfit_em.powerfitter import OPENCL


@skipIf(not OPENCL, "CL not available")
@skipIf(True, "powerfit._powerfit module does not have rotate_image3d function")
class TestRotateImage3d(unittest.TestCase):
    """Test the rotate_image3d function."""

    def setUp(self):
        self.shape = (5, 6, 7)
        self.radius = 3
        self.center = np.asarray([0, 0, 0], dtype=np.float64)
        self.rotmat = np.asarray([1, 0, 0, 0, 1, 0, 0, 0, 1], dtype=np.float64).reshape(
            3, 3
        )
        self.image = np.zeros(self.shape, dtype=np.float64)
        self.image[0, 0, : self.radius] = 1
        self.out = np.zeros_like(self.image)
        self.answer = np.zeros_like(self.image)

    def test_identity(self):
        # simple identity rotation
        _powerfit.rotate_image3d(
            self.image, self.rotmat, self.center, self.radius, self.out
        )
        self.assertTrue(np.allclose(self.out, self.image))

    def test_radius(self):
        # radius too short
        radius = 1
        _powerfit.rotate_image3d(
            self.image, self.rotmat, self.center, self.radius, self.out
        )
        self.answer = self.image
        self.assertTrue(np.allclose(self.answer, self.out))

    def test_center(self):
        self.center = np.asarray((1, 0, 0), dtype=np.float64)
        _powerfit.rotate_image3d(
            self.image, self.rotmat, self.center, self.radius, self.out
        )
        self.answer[0, 0, : self.radius - 1] = 1
        self.answer[0, 0, -1] = 1
        self.assertTrue(np.allclose(self.answer, self.out))

    def test_z_rotation_m90(self):
        # -90 degree z-rotation
        self.rotmat = euler(np.radians(-90), "z")
        _powerfit.rotate_image3d(
            self.image, self.rotmat, self.center, self.radius, self.out
        )
        self.answer[0, -self.radius + 1 :, 0] = 1
        self.answer[0, 0, 0] = 1
        self.assertTrue(np.allclose(self.answer, self.out))

    def test_z_rotation_90(self):
        # +90 degree z-rotation
        self.rotmat = euler(np.radians(90), "z")
        _powerfit.rotate_image3d(
            self.image, self.rotmat, self.center, self.radius, self.out
        )
        self.answer[0, : self.radius, 0] = 1
        self.assertTrue(np.allclose(self.answer, self.out))

    def test_near(self):
        self.rotmat = euler(np.radians(85), "z")
        _powerfit.rotate_image3d(
            self.image, self.rotmat, self.center, self.radius, self.out, nearest=True
        )
        self.answer[0, : self.radius, 0] = 1
        self.assertTrue(np.allclose(self.answer, self.out))


if __name__ == "__main__":
    unittest.main()




from collections import defaultdict, OrderedDict
from collections.abc import Sequence
from io import TextIOWrapper
import operator
from string import capwords

import numpy as np

from .elements import ELEMENTS

# records
MODEL = 'MODEL '
ATOM = 'ATOM  '
HETATM = 'HETATM'
TER = 'TER   '

MODEL_LINE = 'MODEL ' + ' ' * 4 + '{:>4d}\n'
ENDMDL_LINE = 'ENDMDL\n'
TER_LINE = 'TER   ' + '{:>5d}' + ' ' * 6 + '{:3s}' + ' ' + '{:1s}' + \
        '{:>4d}' + '{:1s}' + ' ' * 53 + '\n'
ATOM_LINE = '{:6s}' + '{:>5d}' + ' ' + '{:4s}' + '{:1s}' + '{:3s}' + ' ' + \
        '{:1s}' + '{:>4d}' + '{:1s}' + ' ' * 3 + '{:8.3f}' * 3 + '{:6.2f}' * 2 + \
        ' ' * 10 + '{:<2s}' + '{:2s}\n'
END_LINE = 'END   \n'

ATOM_DATA = ('record id name alt resn chain resi i x y z q b ' \
        'e charge').split()
TER_DATA = 'id resn chain resi i'.split()


def parse_pdb(infile):

    if isinstance(infile, TextIOWrapper):
        f = infile
    elif isinstance(infile, str):
        f = open(infile)
    else:
        raise TypeError('Input should be either a file or string.')

    pdb = defaultdict(list)
    model_number = 1
    for line in f:
        record = line[:6]
        if record in (ATOM, HETATM):
            pdb['model'].append(model_number)
            pdb['record'].append(record)
            pdb['id'].append(int(line[6:11]))
            name = line[12:16].strip()
            pdb['name'].append(name)
            pdb['alt'].append(line[16])
            pdb['resn'].append(line[17:20].strip())
            pdb['chain'].append(line[21])
            pdb['resi'].append(int(line[22:26]))
            pdb['i'].append(line[26])
            pdb['x'].append(float(line[30:38]))
            pdb['y'].append(float(line[38:46]))
            pdb['z'].append(float(line[46:54]))
            pdb['q'].append(float(line[54:60]))
            pdb['b'].append(float(line[60:66]))
            # Be forgiving when determining the element
            e = line[76:78].strip()
            if not e:
                # If element is not given, take the first non-numeric letter of
                # the name as element.
                for e in name:
                    if e.isalpha():
                        break
            pdb['e'].append(e)
            pdb['charge'].append(line[78: 80].strip())
        elif record == MODEL:
            model_number = int(line[10: 14])
    f.close()
    return pdb


def tofile(pdb, out):

    f = open(out, 'w')

    nmodels = len(set(pdb['model']))
    natoms = len(pdb['id'])
    natoms_per_model = natoms // nmodels

    for nmodel in range(nmodels):
        offset = int(nmodel * natoms_per_model)
        # write MODEL record
        if nmodels > 1:
            f.write(MODEL_LINE.format(nmodel + 1))
        prev_chain = pdb['chain'][offset]
        for natom in range(natoms_per_model):
            index = int(offset + natom)

            # write TER record
            current_chain = pdb['chain'][index]
            if prev_chain != current_chain:
                prev_record = pdb['record'][index - 1]
                if prev_record == ATOM:
                    line_data = [pdb[data][index - 1] for data in TER_DATA]
                    line_data[0] += 1
                    f.write(TER_LINE.format(*line_data))
                prev_chain = current_chain

            # write ATOM/HETATM record
            line_data = [pdb[data][index] for data in ATOM_DATA]
            # take care of the rules for atom name position
            e = pdb['e'][index]
            name = pdb['name'][index]
            if len(e) == 1 and len(name) != 4:
                line_data[2] = ' ' + name
            f.write(ATOM_LINE.format(*line_data))

        # write ENDMDL record
        if nmodels > 1:
            f.write(ENDMDL_LINE)

    f.write(END_LINE)
    f.close()


def pdb_dict_to_array(pdb):
    dtype = [('record', np.str_, 6), ('id', np.int32),
             ('name', np.str_, 4), ('alt', np.str_, 1),
             ('resn', np.str_, 4), ('chain', np.str_, 2),
             ('resi', np.int32), ('i', np.str_, 1), ('x', np.float64),
             ('y', np.float64), ('z', np.float64),
             ('q', np.float64), ('b', np.float64),
             ('e', np.str_, 2), ('charge', np.str_, 2),
             ('model', np.int32)]

    natoms = len(pdb['id'])
    pdb_array = np.empty(natoms, dtype=dtype)
    for data in ATOM_DATA:
        pdb_array[data] = pdb[data]
    pdb_array['model'] = pdb['model']
    return pdb_array


def pdb_array_to_dict(pdb_array):
    pdb = defaultdict(list)
    for data in ATOM_DATA:
        pdb[data] = pdb_array[data].tolist()
    pdb['model'] = pdb_array['model'].tolist()
    return pdb


class Structure(object):

    @classmethod
    def fromfile(cls, fid):
        """Initialize Structure from PDB-file"""
        try:
            fname = fid.name
        except AttributeError:
            fname = fid

        if fname[-3:] in ('pdb', 'ent'):
            arr = pdb_dict_to_array(parse_pdb(fid))
        elif fname[-3:] == 'cif':
            arr = mmcif_dict_to_array(parse_mmcif(fid))
        else:
            raise IOError('Filetype not recognized.')
        return cls(arr)

    def __init__(self, pdb):
        self.data = pdb

    @property
    def atomnumber(self):
        """Return array of atom numbers"""
        return self._get_property('number')

    @property
    def chain_list(self):
        return np.unique(self.data['chain'])

    def combine(self, structure):
        return Structure(np.hstack((self.data, structure.data)))

    @property
    def coor(self):
        """Return the coordinates"""
        return np.asarray([self.data['x'], self.data[ 'y'], self.data['z']])

    def duplicate(self):
        """Duplicate the object"""
        return Structure(self.data.copy())

    def _get_property(self, ptype):
        elements, ind = np.unique(self.data['e'], return_inverse=True)
        return np.asarray([getattr(ELEMENTS[capwords(e)], ptype) 
            for e in elements], dtype=np.float64)[ind]

    @property
    def mass(self):
        return self._get_property('mass')

    def rmsd(self, structure):
        return np.sqrt(((self.coor - structure.coor) ** 2).mean() * 3)

    def rotate(self, rotmat):
        """Rotate atoms"""
        self.data['x'], self.data['y'], self.data['z'] = (
              np.asmatrix(rotmat) * np.asmatrix(self.coor)
              )

    def select(self, identifier, values, loperator='==', return_ind=False):
        """A simple way of selecting atoms"""
        if loperator == '==':
            oper = operator.eq
        elif loperator == '<':
            oper = operator.lt
        elif loperator == '>':
            oper = operator.gt
        elif loperator == '>=':
            oper = operator.ge
        elif loperator == '<=':
            oper = operator.le
        elif loperator == '!=':
            oper = operator.ne
        else:
            raise ValueError('Logic operator not recognized.')

        if not isinstance(values, Sequence) or isinstance(values, str):
            values = (values,)

        selection = oper(self.data[identifier], values[0])
        if len(values) > 1:
            for v in values[1:]:
                if loperator == '!=':
                    selection &= oper(self.data[identifier], v)
                else:
                    selection |= oper(self.data[identifier], v)

        if return_ind:
            return selection
        else:
            return Structure(self.data[selection])

    @property
    def sequence(self):
        resids, indices = np.unique(self.data['resi'], return_index=True)
        return self.data['resn'][indices]

    def translate(self, trans):
        """Translate atoms"""
        self.data['x'] += trans[0]
        self.data['y'] += trans[1]
        self.data['z'] += trans[2]

    def tofile(self, fid):
        """Write instance to PDB-file"""
        tofile(pdb_array_to_dict(self.data), fid)

    @property
    def rvdw(self):
        return self._get_property('vdwrad')


def parse_mmcif(infile):
    if isinstance(infile, TextIOWrapper):
        pass
    elif isinstance(infile, str):
        infile = open(infile)
    else:
        raise TypeError("Input should either be a file or string.")

    atom_site = OrderedDict()
    with infile as f:
        for line in f:

            if line.startswith('_atom_site.'):
                words = line.split('.')
                atom_site[words[1].strip()] = []

            if line.startswith('ATOM'):
                words = line.split()
                for key, word in zip(atom_site, words):
                    atom_site[key].append(word)
    return atom_site


def mmcif_dict_to_array(atom_site):

    natoms = len(atom_site['id'])
    dtype = [('record', np.str_, 6), ('id', np.int32),
             ('name', np.str_, 4), ('alt', np.str_, 1),
             ('resn', np.str_, 4), ('chain', np.str_, 2),
             ('resi', np.int32), ('i', np.str_, 1), ('x', np.float64),
             ('y', np.float64), ('z', np.float64),
             ('q', np.float64), ('b', np.float64),
             ('e', np.str_, 2), ('charge', np.str_, 2),
             ('model', np.int32)]

    cifdata = np.zeros(natoms, dtype=dtype)
    cifdata['record'] = 'ATOM  '
    cifdata['id'] = atom_site['id']
    cifdata['name'] = atom_site['label_atom_id']
    cifdata['resn'] = atom_site['label_comp_id']
    cifdata['chain'] = atom_site['label_asym_id']
    cifdata['resi'] = atom_site['label_seq_id']
    cifdata['x'] = atom_site['Cartn_x']
    cifdata['y'] = atom_site['Cartn_y']
    cifdata['z'] = atom_site['Cartn_z']
    cifdata['q'] = atom_site['occupancy']
    cifdata['b'] = atom_site['B_iso_or_equiv']
    cifdata['e'] = atom_site['type_symbol']
    cifdata['charge'] = atom_site['pdbx_formal_charge']
    cifdata['model'] = atom_site['pdbx_PDB_model_num']
    return cifdata



import sys

from numpy.fft import fftn

from powerfit_em import Volume
from powerfit_em._powerfit import fsc_curve

def main():
    vol1 = Volume.fromfile(sys.argv[1])
    vol2 = Volume.fromfile(sys.argv[2])

    ft_vol1 = fftn(vol1.array)
    ft_vol2 = fftn(vol2.array)

    fsc = fsc_curve(ft_vol1, ft_vol2)
    res = [n / vol1.dimensions[0] for n in range(fsc.size)]
    inv_res = [1 / r for r in res[1:]]

    print(fsc)
    print()
    print(res)
    print()
    print(inv_res)



#! ../env/bin/python


from functools import partial
from os.path import splitext, join, abspath
from time import time
from argparse import ArgumentDefaultsHelpFormatter, ArgumentParser, BooleanOptionalAction, FileType
import logging
from typing import BinaryIO, TextIO
import warnings

from rich.logging import RichHandler
from tqdm import TqdmExperimentalWarning
from tqdm.rich import tqdm as rich_tqdm
from tqdm.auto import tqdm

from powerfit_em import (
    __version__,
    Volume,
    Structure,
    structure_to_shape_like,
    proportional_orientations,
    quat_to_rotmat,
    determine_core_indices,
)
from powerfit_em.powerfitter import PowerFitter
from powerfit_em.analyzer import Analyzer
from powerfit_em.helpers import mkdir_p, write_fits_to_pdb, fisher_sigma
from powerfit_em.volume import extend, nearest_multiple2357, trim, resample
try:
    import pyopencl as cl
    OPENCL = True
except:
    OPENCL = False

logger = logging.getLogger(__name__)

warnings.filterwarnings("ignore", category=TqdmExperimentalWarning)


def make_parser():
    """Create the command-line argument parser."""
    p = ArgumentParser(formatter_class=ArgumentDefaultsHelpFormatter)

    p.add_argument(
        "--version", action="version", version=f"%(prog)s {__version__}"
    )

    # Positional arguments
    p.add_argument(
        "target",
        type=FileType("rb"),
        help="Target density map to fit the model in. "
        "Data should either be in CCP4 or MRC format",
    )
    p.add_argument("resolution", type=float, help="Resolution of map in angstrom")
    p.add_argument(
        "template",
        type=FileType("r"),
        help="Atomic model to be fitted in the density. "
        "Format should either be PDB or mmCIF",
    )

    # Optional arguments and flags
    p.add_argument(
        "-a",
        "--angle",
        dest="angle",
        type=float,
        default=10,
        metavar="<float>",
        help="Rotational sampling density in degree. Increasing "
        "this number by a factor of 2 results in approximately "
        "8 times more rotations sampled.",
    )
    # Scoring flags
    p.add_argument(
        "-l",
        "--laplace",
        dest="laplace",
        action="store_true",
        help="Use the Laplace pre-filter density data. "
        "Can be combined "
        "with the core-weighted local cross-correlation.",
    )
    p.add_argument(
        "-cw",
        "--core-weighted",
        dest="core_weighted",
        action="store_true",
        help="Use core-weighted local cross-correlation score. "
        "Can be combined with the Laplace pre-filter.",
    )
    # Resampling
    p.add_argument(
        "-nr",
        "--no-resampling",
        dest="no_resampling",
        action="store_true",
        help="Do not resample the density map.",
    )
    p.add_argument(
        "-rr",
        "--resampling-rate",
        dest="resampling_rate",
        type=float,
        default=2,
        metavar="<float>",
        help="Resampling rate compared to Nyquist.",
    )
    # Trimming related
    p.add_argument(
        "-nt",
        "--no-trimming",
        dest="no_trimming",
        action="store_true",
        help="Do not trim the density map.",
    )
    p.add_argument(
        "-tc",
        "--trimming-cutoff",
        dest="trimming_cutoff",
        type=float,
        default=None,
        metavar="<float>",
        help="Intensity cutoff to which the map will be trimmed. "
        "Default is 10 percent of maximum intensity.",
    )
    # Selection parameter
    p.add_argument(
        "-c",
        "--chain",
        dest="chain",
        type=str,
        default=None,
        metavar="<char>",
        help=(
            "The chain IDs of the structure to be fitted. "
            "Multiple chains can be selected using a comma separated list, i.e. -c A,B,C. "
            "Default is the whole structure."
        ),
    )
    # Output parameters
    p.add_argument(
        "-d",
        "--directory",
        dest="directory",
        type=abspath,
        default=".",
        metavar="<dir>",
        help="Directory where the results are stored.",
    )
    p.add_argument(
        "-n",
        "--num",
        dest="num",
        type=int,
        default=10,
        metavar="<int>",
        help="Number of models written to file. This number "
        "will be capped if less solutions are found as requested.",
    )
    # Computational resources parameters
    p.add_argument(
        "-g",
        "--gpu",
        dest="gpu",
        nargs="?",
        const="0:0",
        default=None,
        metavar="[<platform>:<device>]",
        help="Off-load the intensive calculations to the GPU. Optionally specify platform and device as <platform>:<device> (e.g., --gpu 0:3). If not specified, uses first device in first platform. If omitted, does not use GPU.",
    )
    p.add_argument(
        "-p",
        "--nproc",
        dest="nproc",
        type=int,
        default=1,
        metavar="<int>",
        help="Number of processors used during search. "
        "The number will be capped at the total number "
        "of available processors on your machine.",
    )
    p.add_argument(
        "--log-level",
        dest="log_level",
        type=str,
        default="INFO",
        choices=["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"],
        help="Set the logging level.",
    )
    p.add_argument(
        "--delimiter",
        dest="delimiter",
        type=str,
        default=None,
        metavar="<str>",
        help="Delimiter used in the 'solutions.out' file. For example use ',' or '\\t'. Defaults to fixed width.",
    )
    p.add_argument(
        "--progressbar",
        dest="progressbar",
        action=BooleanOptionalAction,
        default=True,
        help="Show a progress bar during the search. Disabling the progressbar will improve performance.",
    )

    return p

def parse_args():
    """Parse command-line options."""
    p = make_parser()
    args = p.parse_args()

    return args


def get_filetype_template(fname):
    """Determine the file type from the file extension."""
    ext = splitext(fname)[1][1:]
    if ext in ["pdb", "ent"]:
        ft = "pdb"
    elif ext in ["map", "ccp4"]:
        ft = "map"
    else:
        msg = "Filetype of file {:} is not recognized.".format(fname)
        raise IOError(msg)
    return ft

def configure_logging(log_file, log_level= "INFO"):
    for handler in logging.root.handlers:
        logging.root.removeHandler(handler)

    # Write log messages to a file
    if log_file:
        file_handler = logging.FileHandler(log_file)
        file_handler.setLevel(log_level)
        file_handler.setFormatter(logging.Formatter("%(asctime)s %(message)s"))
        logging.root.addHandler(file_handler)
    
    # Write to console with rich formatting
    console_handler = RichHandler(show_time=False, show_path=False, show_level=False)
    console_handler.setLevel(log_level)
    logging.root.addHandler(console_handler)
    logger.setLevel(log_level)


def main():
    args = parse_args()
    
    mkdir_p(args.directory)
    configure_logging(join(args.directory, "powerfit.log"), args.log_level)
    
    progress = partial(
        rich_tqdm, desc="Processing rotations", unit="rot"
    ) if args.progressbar else None

    powerfit(
        target_volume=args.target,
        resolution=args.resolution,
        template_structure=args.template,
        angle=args.angle,
        laplace=args.laplace,
        core_weighted=args.core_weighted,
        no_resampling=args.no_resampling,
        resampling_rate=args.resampling_rate,
        no_trimming=args.no_trimming,
        trimming_cutoff=args.trimming_cutoff,
        chain=args.chain,
        directory=args.directory,
        num=args.num,
        gpu=args.gpu,
        nproc=args.nproc,
        delimiter=args.delimiter,
        progress=progress
    )


def powerfit(target_volume: BinaryIO,
             resolution: float,
             template_structure: TextIO,
             angle: float=10,
             laplace: bool=False,
             core_weighted: bool=False,
             no_resampling: bool=False,
             resampling_rate: float=2,
             no_trimming: bool=False,
             trimming_cutoff: float | None=None,
             chain: str | None =None,
             directory: str='.',
             num: int=10,
             gpu: str | None =None, 
             nproc: int=1,
             delimiter: str = None,
             progress: partial[tqdm] | None = tqdm
             ):
    time0 = time()
    mkdir_p(directory)

    # Get GPU queue if requested
    queues = None
    if gpu:
        if not OPENCL:
            msg = "Running on GPU requires the pyopencl package, however importing pyopencl failed."
            raise ValueError(msg)
        # TODO allow to omit platform, so gpu='4' runs 5th device on first platform
        if isinstance(gpu, str) and ':' in gpu:
            platform_idx, device_idx = map(int, gpu.split(':'))
        else:
            platform_idx, device_idx = 0, 0
        platforms = cl.get_platforms()
        if platform_idx >= len(platforms):
            raise RuntimeError(f"Requested OpenCL platform {platform_idx} not found.")
        platform = platforms[platform_idx]
        devices = platform.get_devices()
        if device_idx >= len(devices):
            raise RuntimeError(f"Requested OpenCL device {device_idx} not found on platform {platform_idx}.")
        context = cl.Context(devices=[devices[device_idx]])
        queues = [cl.CommandQueue(context, device=devices[device_idx])]

    logger.info("Target file read from: {:s}".format(abspath(target_volume.name)))
    target = Volume.fromfile(target_volume)
    logger.info("Target resolution: {:.2f}".format(resolution))
    logger.info(("Initial shape of density:" + " {:d}" * 3).format(*target.shape))
    # Resample target density if requested
    if not no_resampling:
        factor = 2 * resampling_rate * target.voxelspacing / resolution
        if factor < 0.9:
            target = resample(target, factor)
            logger.info(("Shape after resampling:" + " {:d}" * 3).format(*target.shape))
    # Trim target density if requested
    if not no_trimming:
        if trimming_cutoff is None:
            trimming_cutoff = target.array.max() / 10
        target = trim(target, trimming_cutoff)
        logger.info(("Shape after trimming:" + " {:d}" * 3).format(*target.shape))
    # Extend the density to a multiple of 2, 3, 5, and 7 for clFFT
    extended_shape = [nearest_multiple2357(n) for n in target.shape]
    target = extend(target, extended_shape)
    logger.info(("Shape after extending:" + " {:d}" * 3).format(*target.shape))

    # Read in structure or high-resolution map
    logger.info("Template file read from: {:s}".format(abspath(template_structure.name)))
    structure = Structure.fromfile(template_structure)
    if chain is not None:
        logger.info("Selecting chains: " + chain)
        structure = structure.select("chain", chain.split(","))
    if structure.data.size == 0:
        raise ValueError("No atoms were selected.")

    # Move structure to origin of density
    structure.translate(target.origin - structure.coor.mean(axis=1))
    template = structure_to_shape_like(
        target,
        structure.coor,
        resolution=resolution,
        weights=structure.atomnumber,
        shape="vol",
    )
    mask = structure_to_shape_like(
        target, structure.coor, resolution=resolution, shape="mask"
    )

    # Read in the rotations to sample
    logger.info("Reading in rotations.")
    q, w, degree = proportional_orientations(angle)
    rotmat = quat_to_rotmat(q)
    logger.info("Requested rotational sampling density: {:.2f}".format(angle))
    logger.info("Real rotational sampling density: {:}".format(degree))

    # Apply core-weighted mask if requested
    if core_weighted:
        logger.info("Calculating core-weighted mask.")
        mask.array = determine_core_indices(mask.array)

    pf = PowerFitter(target, rotmat, template, mask, queues, laplace=laplace)
    pf._nproc = nproc
    pf.directory = directory
    if gpu:
        logger.info("Using GPU-accelerated search.")
    else:
        logger.info("Requested number of processors: {:d}".format(nproc))
    logger.info("Starting search")
    time1 = time()
    pf.scan(progress=progress)
    dtime = time() - time1
    if dtime < 10:
        logger.info("Time for search: {:.3f} s".format(dtime))
    else:
        logger.info("Time for search: {:.0f}m {:.0f}s".format(*divmod(dtime, 60)))
    logger.info("Analyzing results")
    # calculate the molecular volume of the structure
    mv = (
        structure_to_shape_like(
            target,
            structure.coor,
            resolution=resolution,
            radii=structure.rvdw,
            shape="mask",
        ).array.sum()
        * target.voxelspacing**3
    )
    z_sigma = fisher_sigma(mv, resolution)
    analyzer = Analyzer(
        pf._lcc,
        rotmat,
        pf._rot,
        voxelspacing=target.voxelspacing,
        origin=target.origin,
        z_sigma=z_sigma,
    )

    logger.info("Writing solutions to file.")
    Volume(pf._lcc, target.voxelspacing, target.origin).tofile(
        join(directory, "lcc.mrc")
    )
    analyzer.tofile(join(directory, "solutions.out"), delimiter=delimiter)

    logger.info("Writing PDBs to file.")
    n = min(num, len(analyzer.solutions))
    write_fits_to_pdb(
        structure, analyzer.solutions[:n], basename=join(directory, "fit")
    )

    logger.info("Total time: {:.0f}m {:.0f}s".format(*divmod(time() - time0, 60)))


if __name__ == "__main__":
    main()
